/* sbsmtp - bsmtp sender
 * vix 28mar91 [written]
 */

/* purpose: takes bsmtp script on stdin, destination machine on command line.
 *	vfork/exec "catbat SBSMTPDIR/DESTMACHINE", and feed catbat the bsmtp
 *	script without the HELO and QUIT lines.  we are separate from catbat
 *	because this way, only we need to know what bsmtp is, while catbat
 *	could be used, say, for input nntp batching.
 */

/* portability: none. this is intensely UNIX-specific, will not run on VMS.
 *	could be made to run on non-BSD/OSF with "-Dvfork=fork", I think.
 */

/* cleanliness: none. we reuse variables for things they aren't useful for.
 *	what began as "use only a little bit of memory" got out of hand.
 */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h>

#ifndef CATBAT
#define	CATBAT		"/usr/local/etc/bsmtp/catbat"
#endif

#ifndef SPOOLDIR
#define	SPOOLDIR	"/var/spool/bsmtp"
#endif

#define	DIRMODE		0750
#define LONGEST_LINE	1024

main(argc, argv, envp)
	int argc;
	char *argv[], *envp[];
{
	char	line[LONGEST_LINE], spooldir[MAXPATHLEN];
	int	p[2], child;
	enum	{HELO, DATA, DOT, QUIT} state;
	FILE	*f;
	int	w;

	if (argc != 2) {
		fprintf(stderr, "usage:  %s destinationname\n", argv[0]);
		exit(EX_USAGE);
	}

	umask(0777 ^ DIRMODE);

	sprintf(spooldir, "%s/%s", SPOOLDIR, argv[1]);
	if (rmkdir(spooldir) < 0) {
		perror(spooldir);
		exit(EX_CANTCREAT);
	}

	/* is the first line a HELO?
	 */
	if (!fgets(line, LONGEST_LINE, stdin)) {
		fprintf(stderr, "%s:  empty stdin\n", argv[0]);
		exit(EX_NOINPUT);
	}
	if (strncasecmp(line, "helo ", 5)) {
		fprintf(stderr, "%s:  stdin doesn't begin with HELO (%-.5s)\n",
			argv[0], line);
		exit(EX_DATAERR);
	}
	/* OK, it's a HELO, now ignore it. */

	/* what's the cheapest and most secure popen() in the world?
	 */
	if (pipe(p) < 0) {
		perror("pipe(2)");
		exit(EX_OSERR);
	}
	if (!(child = vfork())) {
		/* child */
		close(p[1]);
		dup2(p[0], 0);
		close(p[0]);
		/* reuse our argv[] */
		argv[0] = CATBAT;
		argv[1] = spooldir;
		argv[2] = NULL;
		execve(argv[0], argv, envp);
		perror("execve(2)");
		_exit(EX_UNAVAILABLE);
	}

	/* parent */
	close(p[0]);
	if (!(f = fdopen(p[1], "w"))) {
		perror("fdopen(2)");
		_exit(EX_OSERR);
	}
	state = HELO;
	while (fgets(line, LONGEST_LINE, stdin)) {
		switch (state) {
		case HELO:
			if (!strncasecmp(line, "data", 4))
				state = DATA;
			break;
		case DATA:
			if (line[0] == '.' && (line[1]=='\n' || line[1]=='\r'))
				state = DOT;
			break;
		case DOT:
			if (!strncasecmp(line, "quit", 4)) {
				state = QUIT;
				/* ignore it. */
				continue;
			}
			break;
		case QUIT:
			/* after a QUIT, nothing else matters. should probably
			 * just break out of the loop, but C makes that hard.
			 */
			continue;
		}
		fputs(line, f);
	}
	fclose(f);
	/* reuse our argc */
	if (child != (argc = wait(&w))) {
		fprintf(stderr, "vfork pid %d, wait pid %d?\n", child, argc);
		exit(EX_OSERR);
	}
	exit(WEXITSTATUS(w));
}

int
strncasecmp(s1, s2, n)
	register u_char *s1, *s2;
	register u_int n;
{
	register u_char c1 = 0, c2 = 0;

	while (n--) {
		if (c1 = *s1++, isupper(c1)) c1 = tolower(c1);
		if (c2 = *s2++, isupper(c2)) c2 = tolower(c2);
		if (c1 != c2) break;
	}
	return c1 - c2;
}

int
rmkdir(d)
	char *d;
{
	extern char *strchr();
	register char *p = d;
	register int r;

	while (*++p) {
		register char *q = strchr(p, '/');
		if (q) { *q = '\0'; }
		r = mkdir(d, DIRMODE);
		if (q) { *q = '/'; p = q; }
		if (r < 0 && errno != EEXIST) {
			perror(d);
			break;
		}
		r = 0;
	}
	return r;
}
